#define STRICT 1
#define ORBITER_MODULE

#pragma comment (lib, "gdiplus.lib")

#include "Orbitersdk.h"
#include <windows.h>
#include <gdiplus.h>
#include <io.h>
#include <stdio.h>
#include <time.h>
#include <comdef.h>

using namespace Gdiplus;
using namespace std;

typedef enum { FORMAT_JPG, FORMAT_BMP } ImageFormat;
#define BTDOWN 0x8000

HWND renderWindow;
DWORD renderWidth;
DWORD renderHeight;
BOOL renderFullscreen;

WNDPROC oldproc;

ImageFormat format = FORMAT_JPG;
ULONG jpgQuality = 100;

bool selectRegion = false;
POINTS selectionStart;
bool startPointSelected = false;

void CaptureScreen(RECT region);
RECT GetAllRegion();

LRESULT APIENTRY MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
		case WM_SYSKEYUP:
			if (wParam == VK_F12) {
				if (GetKeyState(VK_MENU) & BTDOWN) CaptureScreen(GetAllRegion());
			} else if (wParam == VK_F11) {
				if (GetKeyState(VK_MENU) & BTDOWN) selectRegion = true;
			}
			break;
		case WM_LBUTTONDOWN:
			if (selectRegion) {
				selectionStart = MAKEPOINTS(lParam);
				startPointSelected = true;
				return TRUE;
			}
		case WM_LBUTTONUP:
			if (selectRegion) {
				selectRegion = false;
				startPointSelected = false;
				POINTS point = MAKEPOINTS(lParam);
				RECT rect;
				rect.left = selectionStart.x;
				rect.top = selectionStart.y;
				rect.right = point.x;
				rect.bottom = point.y;
				CaptureScreen(rect);
				return TRUE;
			}
		case WM_RBUTTONDOWN:
			if (selectRegion) {
				selectRegion = false;
				startPointSelected = false;
				return TRUE;
			}
		case WM_MOUSEMOVE:
			if (selectRegion && startPointSelected) {
				POINTS point = MAKEPOINTS(lParam);
				RECT rect;
				rect.left = min(selectionStart.x, point.x);
				rect.top = min(selectionStart.y, point.y);
				rect.right = max(selectionStart.x, point.x);
				rect.bottom = max(selectionStart.y, point.y);

				HDC hDC;
				if (renderFullscreen) {
					hDC = GetDC(NULL);
				} else {
					hDC = GetDC(renderWindow);
				}
				DrawFocusRect(hDC, &rect);

				ReleaseDC(NULL, hDC);
				return TRUE;
			}
	}
	return CallWindowProc(oldproc, hWnd, uMsg, wParam, lParam);
}

DLLCLBK void opcOpenRenderViewport (HWND hWnd, DWORD w, DWORD h, BOOL fullscreen)
{
	// load settings
	FILEHANDLE file = oapiOpenFile("ScreenCapture.cfg", FILE_IN, CONFIG);
	if (file != NULL) {
		char strFormat[5];
		if (oapiReadItem_string(file, "format", strFormat)) {
			if (strcmp(strFormat, "bmp") == 0 || strcmp(strFormat, "BMP") == 0) {
				format = FORMAT_BMP;
			} else {
				format = FORMAT_JPG;
			}
		}
		int quality;
		if (oapiReadItem_int(file, "quality", quality)) {
			if (quality < 0) quality = 0;
			if (quality > 100) quality = 100;
			jpgQuality = quality;
		}
		oapiCloseFile(file, FILE_IN);
	}

	// save render window parameters
	renderWindow = hWnd;
	renderWidth = w;
	renderHeight = h;
	renderFullscreen = fullscreen;

	// add render window events processor
	oldproc = (WNDPROC)SetWindowLong(renderWindow, GWL_WNDPROC, (LONG)MsgProc);
}

DLLCLBK void opcCloseRenderViewport() {
	renderWindow = NULL;
}

PBITMAPINFO CreateBitmapInfoStruct(HBITMAP hBmp)
{ 
    BITMAP bmp; 
    PBITMAPINFO pbmi; 
    WORD    cClrBits; 

    // Retrieve the bitmap color format, width, and height. 
	if (!GetObject(hBmp, sizeof(BITMAP), (LPSTR)&bmp)) {
		oapiWriteLog("CreateBitmapInfoStruct: Error in GetObject");
		return NULL;
	}

    // Convert the color format to a count of bits. 
    cClrBits = (WORD)(bmp.bmPlanes * bmp.bmBitsPixel); 
    if (cClrBits == 1) 
        cClrBits = 1; 
    else if (cClrBits <= 4) 
        cClrBits = 4; 
    else if (cClrBits <= 8) 
        cClrBits = 8; 
    else if (cClrBits <= 16) 
        cClrBits = 16; 
    else if (cClrBits <= 24) 
        cClrBits = 24; 
    else cClrBits = 32; 

    // Allocate memory for the BITMAPINFO structure. (This structure 
    // contains a BITMAPINFOHEADER structure and an array of RGBQUAD 
    // data structures.) 

     if (cClrBits != 24) 
         pbmi = (PBITMAPINFO) LocalAlloc(LPTR, 
                    sizeof(BITMAPINFOHEADER) + 
                    sizeof(RGBQUAD) * (1<< cClrBits)); 

     // There is no RGBQUAD array for the 24-bit-per-pixel format. 

     else 
         pbmi = (PBITMAPINFO) LocalAlloc(LPTR, 
                    sizeof(BITMAPINFOHEADER)); 

    // Initialize the fields in the BITMAPINFO structure. 

    pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER); 
    pbmi->bmiHeader.biWidth = bmp.bmWidth; 
    pbmi->bmiHeader.biHeight = bmp.bmHeight; 
    pbmi->bmiHeader.biPlanes = bmp.bmPlanes; 
    pbmi->bmiHeader.biBitCount = bmp.bmBitsPixel; 
    if (cClrBits < 24) 
        pbmi->bmiHeader.biClrUsed = (1<<cClrBits); 

    // If the bitmap is not compressed, set the BI_RGB flag. 
    pbmi->bmiHeader.biCompression = BI_RGB; 

    // Compute the number of bytes in the array of color 
    // indices and store the result in biSizeImage. 
    // For Windows NT, the width must be DWORD aligned unless 
    // the bitmap is RLE compressed. This example shows this. 
    // For Windows 95/98/Me, the width must be WORD aligned unless the 
    // bitmap is RLE compressed.
    pbmi->bmiHeader.biSizeImage = ((pbmi->bmiHeader.biWidth * cClrBits +31) & ~31) /8
                                  * pbmi->bmiHeader.biHeight; 
    // Set biClrImportant to 0, indicating that all of the 
    // device colors are important. 
     pbmi->bmiHeader.biClrImportant = 0; 
     return pbmi; 
 }

bool CreateBMPFile(LPTSTR pszFile, PBITMAPINFO pbi, 
                  HBITMAP hBMP, HDC hDC) { 
    HANDLE hf;                 // file handle 
    BITMAPFILEHEADER hdr;       // bitmap file-header 
    PBITMAPINFOHEADER pbih;     // bitmap info-header 
    LPBYTE lpBits;              // memory pointer 
    DWORD dwTotal;              // total count of bytes 
    DWORD cb;                   // incremental count of bytes 
    BYTE *hp;                   // byte pointer 
    DWORD dwTmp; 

    pbih = (PBITMAPINFOHEADER) pbi; 
    lpBits = (LPBYTE) GlobalAlloc(GMEM_FIXED, pbih->biSizeImage);

	if (!lpBits) {
         oapiWriteLog("CreateBMPFile: Error in GlobalAlloc");
		 return false;
	}

    // Retrieve the color table (RGBQUAD array) and the bits 
    // (array of palette indices) from the DIB. 
    if (!GetDIBits(hDC, hBMP, 0, (WORD) pbih->biHeight, lpBits, pbi, 
        DIB_RGB_COLORS)) {
		oapiWriteLog("CreateBMPFile: Error in GetDIBits");
		return false;
    }

    // Create the .BMP file. 
    hf = CreateFile(pszFile, 
                   GENERIC_READ | GENERIC_WRITE, 
                   (DWORD) 0, 
                    NULL, 
                   CREATE_ALWAYS, 
                   FILE_ATTRIBUTE_NORMAL, 
                   (HANDLE) NULL); 
	if (hf == INVALID_HANDLE_VALUE) {
        oapiWriteLog("CreateBMPFile: Error in CreateFile");
		return false;
	}
    hdr.bfType = 0x4d42;        // 0x42 = "B" 0x4d = "M" 
    // Compute the size of the entire file. 
    hdr.bfSize = (DWORD) (sizeof(BITMAPFILEHEADER) + 
                 pbih->biSize + pbih->biClrUsed 
                 * sizeof(RGBQUAD) + pbih->biSizeImage); 
    hdr.bfReserved1 = 0; 
    hdr.bfReserved2 = 0; 

    // Compute the offset to the array of color indices. 
    hdr.bfOffBits = (DWORD) sizeof(BITMAPFILEHEADER) + 
                    pbih->biSize + pbih->biClrUsed 
                    * sizeof (RGBQUAD); 

    // Copy the BITMAPFILEHEADER into the .BMP file. 
    if (!WriteFile(hf, (LPVOID) &hdr, sizeof(BITMAPFILEHEADER), 
        (LPDWORD) &dwTmp,  NULL)) 
    {
		oapiWriteLog("CreateBMPFile: Error in WriteFile");
		return false;
    }

    // Copy the BITMAPINFOHEADER and RGBQUAD array into the file. 
    if (!WriteFile(hf, (LPVOID) pbih, sizeof(BITMAPINFOHEADER) 
                  + pbih->biClrUsed * sizeof (RGBQUAD), 
				  (LPDWORD) &dwTmp, NULL)) {
        oapiWriteLog("CreateBMPFile: Error in WriteFile2");
		return false;
	}

    // Copy the array of color indices into the .BMP file. 
    dwTotal = cb = pbih->biSizeImage; 
    hp = lpBits; 
	if (!WriteFile(hf, (LPSTR) hp, (int) cb, (LPDWORD) &dwTmp,NULL)) {
		oapiWriteLog("CreateBMPFile: Error in WriteFile3");
		return false;
	}

    // Close the .BMP file. 
	if (!CloseHandle(hf)) {
		oapiWriteLog("CreateBMPFile: Error in CloseHandle");
		return false;
	}

    // Free memory. 
    GlobalFree((HGLOBAL)lpBits);
	return true;
}

int GetEncoderClsid(const WCHAR* format, CLSID* pClsid) {
	UINT  num = 0;          // number of image encoders
	UINT  size = 0;         // size of the image encoder array in bytes

	ImageCodecInfo* pImageCodecInfo = NULL;

	GetImageEncodersSize(&num, &size);
	if(size == 0)
	  return -1;  // Failure

	pImageCodecInfo = (ImageCodecInfo*)(malloc(size));
	if(pImageCodecInfo == NULL)
	  return -1;  // Failure

	GetImageEncoders(num, size, pImageCodecInfo);

	for(UINT j = 0; j < num; ++j)
	{
	  if( wcscmp(pImageCodecInfo[j].MimeType, format) == 0 )
	  {
		 *pClsid = pImageCodecInfo[j].Clsid;
		 free(pImageCodecInfo);
		 return j;  // Success
	  }    
	}

	free(pImageCodecInfo);
	return -1;  // Failure
}

void SaveImage(char* fileName, HBITMAP hBitmap, HDC hDC) {
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);

	Bitmap* bitmap = Bitmap::FromHBITMAP(hBitmap, NULL);
	if (bitmap == NULL) {
		oapiWriteLog("Can't create bitmap");
		return;
	}

	CLSID jpgClsid;
	GetEncoderClsid(L"image/jpeg", &jpgClsid);

	if (&jpgClsid != NULL) {
		EncoderParameters encoderParameters;
		encoderParameters.Count = 1;
		encoderParameters.Parameter[0].Guid = EncoderQuality;
		encoderParameters.Parameter[0].Type = EncoderParameterValueTypeLong;
		encoderParameters.Parameter[0].NumberOfValues = 1;
		encoderParameters.Parameter[0].Value = &jpgQuality;

		_bstr_t str = fileName;
		bitmap->Save(str, &jpgClsid, &encoderParameters);
	}

	delete bitmap;
	GdiplusShutdown(gdiplusToken);
}

void GenerateFileName(char* name) {
	char targetName[50];
	oapiGetObjectName(oapiCameraTarget(), targetName, 50);

	SYSTEMTIME time;
	GetLocalTime(&time);

	string extension = (format == FORMAT_JPG ? "jpg" : "bmp");
	sprintf(name, "(screenshots)\\%02d.%02d.%02d %02d-%02d-%02d %s.%s", time.wYear - 2000, time.wMonth, time.wDay, 
		time.wHour, time.wMinute, time.wSecond, targetName, extension.data());
}

RECT GetAllRegion() {
	RECT rect;
	rect.left = 0;
	rect.top = 0;
	if (renderFullscreen) {
		rect.right = GetSystemMetrics(SM_CXSCREEN);
		rect.bottom = GetSystemMetrics(SM_CYSCREEN);
	} else {
		rect.right = renderWidth;
		rect.bottom = renderHeight;
	}
	return rect;
}

void CaptureScreen(RECT region) {
	int width = abs(region.right - region.left);
	int height = abs(region.bottom - region.top);
	if (width == 0 || height == 0) return;

	HDC hDC;
	if (renderFullscreen) {
		hDC = GetDC(NULL);
	} else {
		hDC = GetDC(renderWindow);
	}

    HDC hMemDC = CreateCompatibleDC(hDC);

	HBITMAP hBitmap = CreateCompatibleBitmap(hDC, width, height);
    if (hBitmap) {
        HBITMAP hOld = (HBITMAP) SelectObject(hMemDC, hBitmap);
		BitBlt(hMemDC, 0, 0, width, height, hDC, 
			min(region.left, region.right), min(region.top, region.bottom), SRCCOPY);
        SelectObject(hMemDC, hOld);
        DeleteDC(hMemDC);

		// check if screenshots directory exists
		if (_access("(screenshots)", 0) == -1) {
			CreateDirectory("(screenshots)", NULL);
		}

		if (format == FORMAT_BMP) {
			PBITMAPINFO bitmapInfo = CreateBitmapInfoStruct(hBitmap);
			if (bitmapInfo != NULL) {
				char name[100];
				GenerateFileName(name);
				CreateBMPFile(name, bitmapInfo, hBitmap, hDC);
			} else {
				oapiWriteLog("Bitmap info creation failed");
			}
		} else {
			char name[100];
			GenerateFileName(name);
			SaveImage(name, hBitmap, hDC);
		}

        ReleaseDC(NULL, hDC);
		DeleteObject(hBitmap);
	}
}
